from typing import List, Optional

from PicImageSearch import Iqdb, Network

from .utils import get_source, handle_img, shorten_url


async def iqdb_search(url: str, proxy: Optional[str], hide_img: bool) -> List[str]:
    async with Network(proxies=proxy) as client:
        iqdb = Iqdb(client=client)
        res = await iqdb.search(url)
        if not res or not res.raw:
            return ["Iqdb 暂时无法使用"]
        selected_res = res.raw[0]
        # 优先取 danbooru 或 yande.re
        danbooru_res_list = [i for i in res.raw if i.source == "Danbooru"]
        yandere_res_list = [i for i in res.raw if i.source == "yande.re"]
        if danbooru_res_list:
            selected_res = danbooru_res_list[0]
        elif yandere_res_list:
            selected_res = yandere_res_list[0]
        thumbnail = await handle_img(selected_res.thumbnail, proxy, hide_img)
        source = (
            await shorten_url(await get_source(selected_res.url, proxy))
            if selected_res.url
            else ""
        )
        res_list = [
            f"Iqdb（{selected_res.similarity}%）",
            thumbnail,
            selected_res.url or "",
            f"来源：{source}" if source else "",
        ]
        return ["\n".join([i for i in res_list if i != ""])]



from typing import List, Optional

from diskcache import Cache

from .config import config


def upsert_cache(cache: Cache, image_md5: str, mode: str, result: List[str]) -> None:
    cache.set(f"{image_md5}_{mode}", result, expire=config.cache_expire * 24 * 60 * 60)


def exist_in_cache(cache: Cache, image_md5: str, mode: str) -> Optional[List[str]]:
    cache_result: Optional[List[str]] = cache.get(f"{image_md5}_{mode}")
    if cache_result:
        cache.touch(f"{image_md5}_{mode}", expire=config.cache_expire * 24 * 60 * 60)
    return cache_result



import itertools
from collections import defaultdict
from typing import Any, Dict, List, Optional, Union

import aiohttp
from lxml.html import HTMLParser, fromstring
from PicImageSearch import EHentai, Network
from PicImageSearch.model import EHentaiItem, EHentaiResponse
from pyquery import PyQuery

from .config import config
from .utils import handle_img, shorten_url


class EHentaiResponseAioHttp:
    def __init__(self, resp_text: str, resp_url: str):
        self.origin: str = resp_text
        utf8_parser = HTMLParser(encoding="utf-8")
        data = PyQuery(fromstring(self.origin, parser=utf8_parser))
        self.raw: List[EHentaiItem] = [
            EHentaiItem(i) for i in data.find(".glcat").parents("tr").items()
        ]
        self.url: str = resp_url


async def ehentai_search(url: str, proxy: Optional[str], hide_img: bool) -> List[str]:
    async with Network(proxies=proxy, cookies=config.exhentai_cookies) as client:
        ex = bool(config.exhentai_cookies)
        ehentai = EHentai(client=client)
        if res := await ehentai.search(url, ex=ex):
            if not res.raw:
                # 如果第一次没找到，使搜索结果包含被删除的部分，并重新搜索
                ehentai = EHentai(client=client, exp=True)
                res = await ehentai.search(url, ex=ex)
            return await search_result_filter(res, proxy, hide_img)
        return ["EHentai 暂时无法使用"]


async def ehentai_title_search(
    title: str, proxy: Optional[str], hide_img: bool
) -> List[str]:
    headers = {}
    if cookies := config.exhentai_cookies:
        headers["Cookie"] = cookies
    url = "https://exhentai.org" if cookies else "https://e-hentai.org"
    params: Dict[str, Any] = {"f_search": title}
    async with aiohttp.ClientSession(headers=headers) as session:
        resp = await session.get(url, proxy=proxy, params=params)
        if res := EHentaiResponseAioHttp(await resp.text(), str(resp.url)):
            if not res.raw:
                # 如果第一次没找到，使搜索结果包含被删除的部分，并重新搜索
                params["advsearch"] = 1
                params["f_sname"] = "on"
                params["f_stags"] = "on"
                params["f_sh"] = "on"
                resp = await session.get(url, proxy=proxy, params=params)
                res = EHentaiResponseAioHttp(await resp.text(), str(resp.url))
            return await search_result_filter(res, proxy, hide_img)
        return ["EHentai 暂时无法使用"]


async def search_result_filter(
    res: Union[EHentaiResponse, EHentaiResponseAioHttp],
    proxy: Optional[str],
    hide_img: bool,
) -> List[str]:
    if not res.raw:
        return ["EHentai 搜索结果为空"]
    # 尽可能过滤掉非预期结果(大概
    priority = defaultdict(lambda: 0)
    priority["Image Set"] = 1
    priority["Non-H"] = 2
    priority["Western"] = 3
    priority["Misc"] = 4
    priority["Cosplay"] = 5
    priority["Asian Porn"] = 6
    res.raw.sort(key=lambda x: priority[x.type], reverse=True)
    for key, group in itertools.groupby(res.raw, key=lambda x: x.type):
        group_list = list(group)
        if priority[key] > 0 and len(res.raw) != len(group_list):
            res.raw = [i for i in res.raw if i not in group_list]

    # 优先找汉化版或原版
    if chinese_res := [
        i for i in res.raw if all(tag in i.tags for tag in ["translated", "chinese"])
    ]:
        selected_res = chinese_res[0]
    elif not_translated_res := [i for i in res.raw if "translated" not in i.tags]:
        selected_res = not_translated_res[0]
    else:
        selected_res = res.raw[0]

    thumbnail = await handle_img(
        selected_res.thumbnail, proxy, hide_img, config.exhentai_cookies
    )
    _url = await shorten_url(res.url)
    res_list = [
        "EHentai 搜索结果",
        thumbnail,
        selected_res.title,
        f"类型：{selected_res.type}",
        f"日期：{selected_res.date}",
        f"来源：{selected_res.url}",
        f"搜索页面：{_url}",
    ]
    return ["\n".join([i for i in res_list if i != ""])]



from typing import List, Optional

from PicImageSearch import Ascii2D, Network
from PicImageSearch.model import Ascii2DResponse

from .utils import handle_img, shorten_url


async def ascii2d_search(url: str, proxy: Optional[str], hide_img: bool) -> List[str]:
    async with Network(proxies=proxy) as client:
        ascii2d_color = Ascii2D(client=client)
        ascii2d_bovw = Ascii2D(bovw=True, client=client)
        color_res = await ascii2d_color.search(url)
        if not color_res or not color_res.raw:
            return ["Ascii2D 暂时无法使用"]
        # 针对 QQ 图片链接反盗链做处理
        if color_res.raw[0].hash == "5bb9bec07e71ef10a1a47521d396811d":
            url = f"https://images.weserv.nl/?url={url}"
            color_res = await ascii2d_color.search(url)
            bovw_res = await ascii2d_bovw.search(url)
        else:
            bovw_res = await ascii2d_bovw.search(url)

        async def get_final_res(res: Ascii2DResponse) -> List[str]:
            if not res.raw[0].url:
                res.raw[0] = res.raw[1]
            thumbnail = await handle_img(res.raw[0].thumbnail, proxy, hide_img)
            _url = await shorten_url(res.raw[0].url) if res.raw[0] else ""
            res_list = [
                thumbnail,
                res.raw[0].title or "",
                f"作者：{res.raw[0].author}" if res.raw[0].author else "",
                f"来源：{_url}" if _url else "",
                f"搜索页面：{res.url}",
            ]
            return [i for i in res_list if i != ""]

        color_final_res = await get_final_res(color_res)
        bovw_final_res = await get_final_res(bovw_res)
        if color_final_res[:-1] == bovw_final_res[:-1]:
            return ["Ascii2D 色合検索与特徴検索結果完全一致\n" + "\n".join(color_final_res)]

        return [
            f"Ascii2D 色合検索結果\n" + "\n".join(color_final_res),
            f"Ascii2D 特徴検索結果\n" + "\n".join(bovw_final_res),
        ]



import asyncio
import re
from collections import defaultdict
from typing import DefaultDict, List, Optional, Tuple, Union

import arrow
from diskcache import Cache
from nonebot.adapters.onebot.v11 import (
    ActionFailed,
    Bot,
    GroupMessageEvent,
    Message,
    MessageEvent,
    PrivateMessageEvent,
)
from nonebot.matcher import Matcher
from nonebot.params import CommandArg
from nonebot.plugin.on import on_command, on_message
from nonebot.rule import Rule

from .ascii2d import ascii2d_search
from .cache import exist_in_cache, upsert_cache
from .config import config
from .ehentai import ehentai_search
from .iqdb import iqdb_search
from .saucenao import saucenao_search
from .utils import handle_img

sending_lock: DefaultDict[Tuple[Union[int, str], str], asyncio.Lock] = defaultdict(
    asyncio.Lock
)


def has_images(event: MessageEvent) -> bool:
    message = event.reply.message if event.reply else event.message
    return bool([i for i in message if i.type == "image"])


async def to_me_with_images(bot: Bot, event: MessageEvent) -> bool:
    at_me = bool(
        [i for i in event.message if i.type == "at" and i.data["qq"] == bot.self_id]
    )
    has_image = has_images(event)
    if isinstance(event, PrivateMessageEvent):
        return has_image and config.search_immediately
    # 群里回复机器人发送的消息时，必须带上 "再搜"才会搜图，否则会被无视
    if event.reply and event.reply.sender.user_id == int(bot.self_id):
        return (
            has_image
            and (event.to_me or at_me)
            and "再搜" in event.message.extract_plain_text()
        )
    return has_image and (event.to_me or at_me)


IMAGE_SEARCH = on_message(rule=Rule(to_me_with_images), priority=5)
IMAGE_SEARCH_MODE = on_command("搜图", priority=5)


@IMAGE_SEARCH_MODE.handle()
async def handle_first_receive(matcher: Matcher, args: Message = CommandArg()) -> None:
    mode, purge = get_args(args)
    matcher.state["ARGS"] = (mode, purge)
    if [i for i in args if i.type == "image"]:
        matcher.set_arg("IMAGES", args)


async def image_search(
    url: str,
    mode: str,
    purge: bool,
    _cache: Cache,
    proxy: Optional[str] = config.proxy,
    hide_img: bool = config.hide_img,
) -> List[str]:
    url = get_universal_img_url(url)
    image_md5 = re.search("[A-F0-9]{32}", url)[0]  # type: ignore
    if not purge and (result := exist_in_cache(_cache, image_md5, mode)):
        return [f"[缓存] {i}" for i in result]
    try:
        if mode == "a2d":
            result = await ascii2d_search(url, proxy, hide_img)
        elif mode == "iqdb":
            result = await iqdb_search(url, proxy, hide_img)
        elif mode == "ex":
            result = await ehentai_search(url, proxy, hide_img)
        else:
            result = await saucenao_search(url, mode, proxy, hide_img)
    except Exception as e:
        thumbnail = await handle_img(url, proxy, False)
        return [f"{thumbnail}\n❌️ 该图搜图失败，请稍后再试\nE: {repr(e)}"]
    upsert_cache(_cache, image_md5, mode, result)
    return result


def get_universal_img_url(url: str) -> str:
    fianl_url = url.replace(
        "/c2cpicdw.qpic.cn/offpic_new/", "/gchat.qpic.cn/gchatpic_new/"
    )
    final_url = re.sub(r"/\d+/+\d+-\d+-", "/0/0-0-", fianl_url)
    final_url = re.sub(r"\?.*$", "", final_url)
    return final_url


def get_image_urls(event: MessageEvent) -> List[str]:
    message = event.reply.message if event.reply else event.message
    return [i.data["url"] for i in message if i.type == "image" and i.data.get("url")]


def get_args(msg: Message) -> Tuple[str, bool]:
    mode = "all"
    plain_text = msg.extract_plain_text()
    args = ["pixiv", "danbooru", "doujin", "anime", "a2d", "ex", "iqdb"]
    if plain_text:
        for i in args:
            if f"--{i}" in plain_text:
                mode = i
                break
    purge = "--purge" in plain_text
    return mode, purge


async def send_result_message(
    bot: Bot, event: MessageEvent, msg_list: List[str]
) -> None:
    if isinstance(event, GroupMessageEvent):
        current_sending_lock = sending_lock[(event.group_id, "group")]
    else:
        current_sending_lock = sending_lock[(event.user_id, "private")]
    if flag := (config.forward_search_result and len(msg_list) > 1):
        try:
            start_time = arrow.now()
            async with current_sending_lock:
                await send_forward_msg(bot, event, msg_list)
                await asyncio.sleep(
                    max(1 - (arrow.now() - start_time).total_seconds(), 0)
                )
        except ActionFailed:
            flag = False
    if not flag:
        for msg in msg_list:
            start_time = arrow.now()
            async with current_sending_lock:
                await send_msg(bot, event, msg)
                await asyncio.sleep(
                    max(1 - (arrow.now() - start_time).total_seconds(), 0)
                )


async def send_msg(bot: Bot, event: MessageEvent, message: str) -> None:
    await bot.send_msg(
        user_id=event.user_id if isinstance(event, PrivateMessageEvent) else 0,
        group_id=event.group_id if isinstance(event, GroupMessageEvent) else 0,
        message=message,
    )


async def send_forward_msg(bot: Bot, event: MessageEvent, msg_list: List[str]) -> None:
    await bot.send_forward_msg(
        user_id=event.user_id if isinstance(event, PrivateMessageEvent) else 0,
        group_id=event.group_id if isinstance(event, GroupMessageEvent) else 0,
        messages=[
            {
                "type": "node",
                "data": {
                    "name": "\u200b",
                    "uin": bot.self_id,
                    "content": msg,
                },
            }
            for msg in msg_list
        ],
    )


@IMAGE_SEARCH.handle()
@IMAGE_SEARCH_MODE.got("IMAGES", prompt="请发送图片")
async def handle_image_search(bot: Bot, event: MessageEvent, matcher: Matcher) -> None:
    image_urls = get_image_urls(event)
    if not image_urls:
        await IMAGE_SEARCH_MODE.reject()
    if "ARGS" in matcher.state:
        mode, purge = matcher.state["ARGS"]
    else:
        mode, purge = get_args(event.message)
    with Cache("picsearch_cache") as c:
        for i in image_urls:
            await send_result_message(bot, event, await image_search(i, mode, purge, c))
        c.expire()
